﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CG2.Mathematics;
using CG2.Rendering;
using CG2.Shading;

namespace CG2.Modeling
{
	public class Mesh : Model
	{
		public class Vertex
		{
			public Vector4 Position;
			// Normal, Color, UV ...

			public Vertex()
			{
			}

			public Vertex(Vector4 position)
			{
				Position = position;
			}

			public Vertex(Double x, Double y, Double z)
			{
				Position = new Vector4(x, y, z);
			}
		}

		public class Face
		{
			public Vertex V0, V1, V2;

			public void Set(Vertex v0, Vertex v1, Vertex v2)
			{
				V0 = v0;
				V1 = v1;
				V2 = v2;

				u = V1.Position - V0.Position;
				v = V2.Position - v0.Position;
				uu = u * u;
				vv = v * v;
				uv = u * v;
				k = 1 / (uv * uv - uu * vv);
				n = (u % v).Normalized;
			}

			public void Colide(Ray ray, Model model)
			{
				Double nv = n * ray.Direction;
				if (nv >= -0.00001 && nv <= 0.00001) return;

				Double param = (n * (V0.Position - ray.Origin)) / nv; // n^T (o_p - o_r) / n^T v

				if (param < Ray.Bias || param >= ray.HitParam) return; // collisions beyond current ray param are ignored

				Vector4 p = ray.Origin + param * ray.Direction;
				Vector4 w = p - V0.Position;
				Double wu = w * u;
				Double wv = w * v;
				Double s = k * (uv * wv - vv * wu);
				Double t = k * (uv * wu - uu * wv);

				if (s < 0 || t < 0 || s + t > 1) return; // collision point p is out of triangle

				ray.HitModel = model;
				ray.HitParam = param;
				ray.HitNormal = n;
				// ## UV from (s,t)
			}

			private Vector4 u, v, n;
			private Double uu, vv, uv, k;
		}


		#region Properties

		public List<Vertex> Vertices = new List<Vertex>();
		public List<Face> Faces = new List<Face>();

		#endregion


		#region Init

		public Mesh()
		{
		}

		public Mesh(Shader shader)
		{
			Shader = shader;
		}

		public void SetVertices(params Vertex[] vertices)
		{
			Vertices = new List<Vertex>(vertices);
		}

		public void SetFacesByIndices(params Int32[] indices)
		{
			Faces = new List<Face>();
			for (Int32 i = 0; i < indices.Length; i += 3)
			{
				Face face = new Face();
				face.Set(Vertices[indices[i]], Vertices[indices[i + 1]], Vertices[indices[i + 2]]);
				Faces.Add(face);
			}
		}

		public void SetAABB(Vector4 min, Vector4 max)
		{
			SetVertices(
				new Vertex(min.X, min.Y, min.Z), new Vertex(min.X, max.Y, min.Z), new Vertex(max.X, max.Y, min.Z), new Vertex(max.X, min.Y, min.Z),
				new Vertex(min.X, min.Y, max.Z), new Vertex(min.X, max.Y, max.Z), new Vertex(max.X, max.Y, max.Z), new Vertex(max.X, min.Y, max.Z)
			);
			SetFacesByIndices(
				0, 4, 5, 0, 5, 1, 3, 2, 6, 3, 6, 7, // X
				0, 3, 7, 7, 4, 0, 1, 5, 6, 6, 2, 1, // Y
				0, 1, 2, 0, 2, 3, 4, 7, 6, 4, 6, 5 // Z
			);
		}

		public void SetPyramid(Vector4 a, Vector4 b, Vector4 c, Vector4 d)
		{
			SetVertices(new Vertex(a), new Vertex(b), new Vertex(c), new Vertex(d));
			SetFacesByIndices(0, 2, 1, 0, 1, 3, 1, 2, 3, 2, 0, 3);
		}


		#endregion


		#region Raytracing

		public override void Collide(Ray ray)
		{
			foreach (Face face in Faces)
			{
				face.Colide(ray, this);
			}
		}

		#endregion
	}
}
