﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CG2.Mathematics;
using CG2.Modeling;

namespace CG2.Rendering
{
	public class BspNode
	{
		#region Properties

		public static Random Rnd = new Random();

		public Vector4 Origin;
		public Vector4 Normal;

		public BspNode UpChild;
		public BspNode DownChild;

		public List<Model> Models;

		#endregion

		public void Init(List<Sphere> spheres)
		{
			if (spheres.Count < 3) // leaf node
			{
				Models = new List<Model>(spheres);
				return;
			}

			Origin = Vector4.Zero;
			foreach (Sphere sphere in spheres)
			{
				Origin += sphere.Origin;
			}
			Origin *= 1.0 / spheres.Count; // average position

			Normal = new Vector4(Rnd.NextDouble(), Rnd.NextDouble(), Rnd.NextDouble());
			Double nl = Normal.Length;
			if (nl == 0) Normal = new Vector4(1, 0, 0);
			else Normal *= 1 / nl; // normalization

			List<Sphere> upSpheres = new List<Sphere>();
			List<Sphere> downSpheres = new List<Sphere>();

			foreach (Sphere sphere in spheres)
			{
				Double noo = Normal * (sphere.Origin - Origin); // sphere plane signed distance (projection on plane normal)
				Boolean isNear = Math.Abs(noo) < sphere.Radius;
				if (isNear || noo > 0) upSpheres.Add(sphere);
				if (isNear || noo < 0) downSpheres.Add(sphere);
			}

			if (upSpheres.Count == spheres.Count || downSpheres.Count == spheres.Count) // leaf node as well
			{
				Models = new List<Model>(spheres);
				return;
			}

			UpChild = new BspNode();
			UpChild.Init(upSpheres);

			DownChild = new BspNode();
			DownChild.Init(downSpheres);
		}

		public void Collide(Ray ray, Double t0, Double t1)
		{
			if (Models != null) // leaf node
			{
				foreach (Model model in Models)
				{
					if (ray == model.Ray) continue;
					model.Ray = ray; // flag model with this ray
					model.Collide(ray);
				}
				return;
			}

			Double nv = Normal * ray.Direction;
			Double noo = Normal * (Origin - ray.Origin);
			if (nv == 0)
			{
				if (noo < 0) UpChild.Collide(ray, t0, t1);
				else DownChild.Collide(ray, t0, t1);
				return;
			}

			Double t = noo / nv;

			if (t < t0) t = t0;
			if (t > t1) t = t1;

			if (nv > 0 && t < t1) UpChild.Collide(ray, t, t1);
			if (nv < 0 && t > t0) UpChild.Collide(ray, t0, t);

			if (nv > 0 && t > t0) DownChild.Collide(ray, t0, t);
			if (nv < 0 && t < t1) DownChild.Collide(ray, t, t1);
		}
	}

	public class BspAccelerator : RayAccelerator
	{
		#region Properties
		
		public BspNode Root;

		#endregion

		public override void BuildStructure(World world)
		{
			World = world;

			List<Sphere> spheres = new List<Sphere>();
			foreach (Model model in world.Models)
			{
				if (model is Sphere) spheres.Add(model as Sphere);
			}

			Root = new BspNode();
			Root.Init(spheres);
		}

		public override void Collide(Ray ray)
		{
			Root.Collide(ray, Ray.Bias, ray.HitParam);
		}
	}
}
