﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CG2.Mathematics;
using System.Drawing;
using CG2.Modeling;
using CG2.Lighting;
using CG2.Shading;

namespace CG2.Rendering
{
	public class Camera
	{
		#region Properties

		public Vector4 Position;
		public Vector4 Target;
		public Vector4 Up = new Vector4(0, 0, 1);
		public Double FovY = 45;
		public Vector4 U, V, W; // camera to world space

		public Vector4 BgColor = new Vector4(0, 0, 0, 0);
		public Model Vacuum = new Model() { Shader = new PhongShader() };

		public RayAccelerator Accelerator = new RayAccelerator();
		public World World;

		public Bitmap Bitmap;
		// TODO: declare native size of Bitmap (BitmapWidth, BitmapHeight)

		public Vector4[] Pixels;
		protected Int32 Width;
		protected Int32 Height;

		public Boolean UseShadows = true;

		public Double MinFactor = 0.01;
		public Int32 ReflectionCount = 0;
		public Int32 RefractionCount = 0;

		public Boolean UseFresnel = true;
		// TODO: declare FresnelPower (=1) being the power coefficient in the Schlick approximation of Fresnel term

		// TODO: declare SubSampleCount (=1) being the number of sub sampling rays per pixel

		#endregion


		#region Init

		public Camera(Int32 bitmapWidth, Int32 bitmapHeight)
		{
			//TODO: init BitmapWidth and BitmapHeight with given parameters
			Width = bitmapWidth;
			Height = bitmapHeight;

			// TODO: Create bitmap in Init buffers instead
			Int32 pixelCount = Width * Height;
			Bitmap = new Bitmap(Width, Height);
			Pixels = new Vector4[pixelCount];
		}

		#endregion


		#region Pixels Access

		public Vector4 GetPixel(Int32 i, Int32 j)
		{
			return Pixels[i + j * Width];
		}

		public void SetPixel(Int32 i, Int32 j, Vector4 color)
		{
			Pixels[i + j * Width] = color;
		}

		#endregion


		#region Rendering

		public void Render()
		{
			Accelerator.BuildStructure(World);

			InitBuffers();

			RayTrace();

			PresentFrame(); // Antialiasing inside
		}

		public void InitBuffers()
		{
			// TODO: set total number of pixels (Width, Height) in Pixels buffer based on the SubSampleCount and size of bitmap
		}

		public void RayTrace()
		{
			W = Position - Target;
			U = Up % W;
			V = W % U;

			U = U.Normalized;
			V = V.Normalized;
			W = W.Normalized;

			Double x0 = (1.0 - Width) / 2;
			Double y0 = (1.0 - Height) / 2;
			Double z = Height / (2 * Math.Tan(MathEx.DegToRad(FovY) / 2));

			Double y = y0;
			for (Int32 j = Height - 1; j >= 0; j--, y += 1.0f)
			{
				Double x = x0;
				for (Int32 i = 0; i < Width; i++, x += 1.0f)
				{
					Vector4 rayDir = (x * U + y * V - z * W).Normalized;
					Ray ray = new Ray(Position, rayDir, Vacuum, 0, 0);
					Vector4 color = RayTrace(ray, 1.0);

					SetPixel(i, j, color);
				}
			}
		}

		public Vector4 RayTrace(Ray ray, Double factor)
		{
			Vector4 color = Vector4.Zero;

			// Intersection detection
			Accelerator.Collide(ray);

			if (ray.HitModel == null) return factor * BgColor; // no collision return

			Vector4 hitPoint = ray.GetHitPoint();
			Vector4 hitNormal = ray.HitNormal;
			Model outModel = ray.HitModel;

			hitNormal = ray.HitModel.Shader.GetNormal(hitPoint, ray.HitNormal);

			if (ray.HitModel == ray.InModel) // exiting transparent model
			{
				hitNormal = -hitNormal;
				outModel = Vacuum; // we exit back into vacuum
			}

			// Lighting
			Ray lightRay = new Ray();
			Vector4 viewDir = -ray.Direction; // hit view vector
			foreach (Light light in World.Lights)
			{
				light.SetLightRayAt(hitPoint, lightRay);

				Double nl = ray.HitNormal * lightRay.Direction;
				if (UseShadows && nl > 0.0f) Accelerator.Collide(lightRay);
				Double intensityFactor = (lightRay.HitModel != null) ? lightRay.HitModel.Shader.RefractionFactor : 1; // due to shadow

				color += ray.HitModel.Shader.GetColor(hitPoint, hitNormal, viewDir, lightRay.Direction, intensityFactor * light.GetIntensityAt(hitPoint), light);
			}

			Double reflectionFactor = ray.HitModel.Shader.ReflectionFactor * factor;
			Double refractionFactor = ray.HitModel.Shader.RefractionFactor * factor;
			Double n1 = ray.InModel.Shader.RefractionIndex;
			Double n2 = outModel.Shader.RefractionIndex;

			// Fresnel (Schlick approximation)
			if (UseFresnel)
			{
				// TODO: modify reflectionFactor and refractionFactor with Fresnel term (use Schlick approximation)
			}

			// Refraction
			Boolean trueRefraction = true;
			if (ray.RefractionLevel < RefractionCount && refractionFactor > MinFactor)
			{
				Vector4 refractedVector;
				trueRefraction = Vector4.Refract(hitNormal, ray.Direction, n1/n2, out refractedVector);
				Model refractedModel = (trueRefraction) ? outModel : ray.InModel;
				Ray refractedRay = new Ray(hitPoint, refractedVector, refractedModel, ray.ReflectionLevel, ray.RefractionLevel + 1);
				color += RayTrace(refractedRay, refractionFactor);
			}

			// Reflection
			if (ray.ReflectionLevel < ReflectionCount && reflectionFactor > MinFactor && trueRefraction)
			{
				Vector4 reflectedVetor;
				Vector4.Reflect(hitNormal, ray.Direction, out reflectedVetor);
				Ray reflectedRay = new Ray(hitPoint, reflectedVetor, ray.InModel, ray.ReflectionLevel + 1, ray.RefractionLevel);
				color += RayTrace(reflectedRay, reflectionFactor);
			}

			return factor * color;
		}

		public void PresentFrame()
		{
			for (Int32 j = 0; j < Height; j++)
			{
				for (Int32 i = 0; i < Width; i++)
				{
					// TODO: calculated subsampled color by averaging colors of subpixels
					Vector4 c = GetPixel(i, j);
					c = Vector4.Clamp(c, 0, 1);

					Bitmap.SetPixel(i, j, Color.FromArgb(
						(Int32)(255 * c.X),
						(Int32)(255 * c.Y),
						(Int32)(255 * c.Z))
					);
				}
			}
		}

		#endregion
	}
}
