﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using CG2.Modeling;
using CG2.Mathematics;

namespace CG2.Rendering
{
	public struct GridCell
	{
		public List<Model> Models;

		public void Add(Model model)
		{
			if (Models == null) Models = new List<Model>();
			Models.Add(model);
		}
	}

	public class GridAccelerator : RayAccelerator
	{
		#region Properties

		public Vector4 Origin;
		public Vector4 CellSize;
		public Vector4 CellCount;

		public GridCell[] Cells;

		#endregion

		public override void BuildStructure(World world)
		{
			World = world;

			Vector4 min = new Vector4(Double.MaxValue, Double.MaxValue, Double.MaxValue);
			Vector4 max = new Vector4(Double.MinValue, Double.MinValue, Double.MinValue);
			Double radius = 0;
			Int32 count = 0;

			foreach (Model model in world.Models)
			{
				if (!(model is Sphere)) continue; // process only spheres

				Sphere sphere = model as Sphere;
				Double r = sphere.Radius;
				Vector4 c = sphere.Origin;

				Vector4 smin = new Vector4(c.X - r, c.Y - r, c.Z - r);
				Vector4 smax = new Vector4(c.X + r, c.Y + r, c.Z + r);

				if (smin.X < min.X) min.X = smin.X;
				if (smin.Y < min.Y) min.Y = smin.Y;
				if (smin.Z < min.Z) min.Z = smin.Z;

				if (smax.X > max.X) max.X = smax.X;
				if (smax.Y > max.Y) max.Y = smax.Y;
				if (smax.Z > max.Z) max.Z = smax.Z;

				radius += r;
				count += 1;
			}

			radius /= count;
			Double cs = 3.0 * radius;
			
			CellSize = new Vector4(cs, cs, cs);
			Origin = min - CellSize;
			CellCount = new Vector4(
				(Int32)((max.X - min.X) / CellSize.X + 3),
				(Int32)((max.Y - min.Y) / CellSize.Y + 3),
				(Int32)((max.Z - min.Z) / CellSize.Z + 3));

			Cells = new GridCell[(Int32)(CellCount.X * CellCount.Y * CellCount.Z)];

			foreach (Model model in world.Models)
			{
				if (!(model is Sphere)) continue; // process only spheres

				Sphere sphere = model as Sphere;
				Double r = sphere.Radius;
				Vector4 c = sphere.Origin;

				Vector4 cmin = new Vector4(
					(Int32)((c.X - r - Origin.X) / CellSize.X),
					(Int32)((c.Y - r - Origin.Y) / CellSize.Y),
					(Int32)((c.Z - r - Origin.Z) / CellSize.Z));

				Vector4 cmax = new Vector4(
					(Int32)((c.X + r - Origin.X) / CellSize.X),
					(Int32)((c.Y + r - Origin.Y) / CellSize.Y),
					(Int32)((c.Z + r - Origin.Z) / CellSize.Z));

				Int32 sy = 1 * (Int32)CellCount.X;
				Int32 sz = sy * (Int32)CellCount.Y;

				for (Int32 cz = (Int32)cmin.Z; cz <= (Int32)cmax.Z; cz++)
				{
					for (Int32 cy = (Int32)cmin.Y; cy <= (Int32)cmax.Y; cy++)
					{
						for (Int32 cx = (Int32)cmin.X; cx <= (Int32)cmax.X; cx++)
						{
							Cells[cx + sy * cy + sz * cz].Add(model);
						}
					}
				}
			}
		}

		public override void Collide(Ray ray)
		{
			// ** Tutorial: ray in grid traversal
			// - Find ray parameters where the ray enters and exits the whole grid (t0, t1)
			// - Clamp both t0 and t1 into (Ray.Bias, ray.HitParam) interval
			// - If t0 > t1 no intersection -> exit
			// - Find the first cell coordinates (cx,cy,cz) where the ray enters the grid (at t0)
			// - Find ray params (tx, ty, tz) where it intersecs first cell boundary in each axis
			// - Find ray param increments (dtx, dty, dtz) when the ray travels to neighboring cell boundary
			// - The grid traversal loop proceeds while t0 < t1
			// - In each iteration first find actual cell at linear index (cx + cy*CellSize.Y + cz*CellSize.Y*CellSize.Z)
			// - for each model in current cell collide ray with object (sphere)
			// - if there was some hit -> we are done -> exit
			// - if we are curretly hitting x-axis (t0 == tx) increment tx to next cell
			// - do the same for y and z
			// - finaly update current ray param t0 to the minimal boundary ray param (tx,ty,tz)

			Double tMinX = (Origin.X + 0 * CellSize.X - ray.Origin.X) / ray.Direction.X;
			Double tMinY = (Origin.Y + 0 * CellSize.Y - ray.Origin.Y) / ray.Direction.Y;
			Double tMinZ = (Origin.Z + 0 * CellSize.Z - ray.Origin.Z) / ray.Direction.Z;
			Double tMaxX = (Origin.X + (CellCount.X - 1) * CellSize.X - ray.Origin.X) / ray.Direction.X;
			Double tMaxY = (Origin.Y + (CellCount.Y - 1) * CellSize.Y - ray.Origin.Y) / ray.Direction.Y;
			Double tMaxZ = (Origin.Z + (CellCount.Z - 1) * CellSize.Z - ray.Origin.Z) / ray.Direction.Z;

			if (tMinX > tMaxX) { MathEx.Swap(ref tMinX, ref tMaxX); }
			if (tMinY > tMaxY) { MathEx.Swap(ref tMinY, ref tMaxY); }
			if (tMinZ > tMaxZ) { MathEx.Swap(ref tMinZ, ref tMaxZ); }

			Double t0 = MathEx.Max3(tMinX, tMinY, tMinZ);
			Double t1 = MathEx.Min3(tMaxX, tMaxY, tMaxZ);

			if (t0 < Ray.Bias) t0 = Ray.Bias;
			if (t1 > ray.HitParam) t1 = ray.HitParam;
			if (t0 > t1) return; // no hit with grid

			Int32 cX = (Int32)((ray.Origin.X + t0 * ray.Direction.X - Origin.X) / CellSize.X);
			Int32 cY = (Int32)((ray.Origin.Y + t0 * ray.Direction.Y - Origin.Y) / CellSize.Y);
			Int32 cZ = (Int32)((ray.Origin.Z + t0 * ray.Direction.Z - Origin.Z) / CellSize.Z);

			Int32 dcX = (ray.Direction.X > 0) ? +1 : -1;
			Int32 dcY = (ray.Direction.Y > 0) ? +1 : -1;
			Int32 dcZ = (ray.Direction.Z > 0) ? +1 : -1;
			Int32 qX = (ray.Direction.X > 0) ? +1 : 0;
			Int32 qY = (ray.Direction.Y > 0) ? +1 : 0;
			Int32 qZ = (ray.Direction.Z > 0) ? +1 : 0;

			Double tX = (Origin.X + (cX + qX) * CellSize.X - ray.Origin.X) / ray.Direction.X;
			Double tY = (Origin.Y + (cY + qY) * CellSize.Y - ray.Origin.Y) / ray.Direction.Y;
			Double tZ = (Origin.Z + (cZ + qZ) * CellSize.Z - ray.Origin.Z) / ray.Direction.Z;
			Double dtX = CellSize.X / Math.Abs(ray.Direction.X);
			Double dtY = CellSize.Y / Math.Abs(ray.Direction.Y);
			Double dtZ = CellSize.Z / Math.Abs(ray.Direction.Z);

			Int32 sy = 1 * (Int32)CellCount.X;
			Int32 sz = sy * (Int32)CellCount.Y;

			while (t0 < t1)
			{
				List<Model> models = Cells[cX + sy * cY + sz * cZ].Models;
				if (models != null)
				{
					foreach (Model model in models)
					{
						if (ray == model.Ray) continue;
						model.Ray = ray; // flag model with this ray
						model.Collide(ray);
					}

					if (ray.HitModel != null) break;
				}

				if (t0 == tX) { cX += dcX; tX += dtX; }
				else if (t0 == tY) { cY += dcY; tY += dtY; }
				else if (t0 == tZ) { cZ += dcZ; tZ += dtZ; }
				t0 = MathEx.Min3(tX, tY, tZ);
			}
		}

	}
}
